# Chopok
# Ski lift statistic simulation
# Copyright (c) 2010 Petr Koupy <petr.koupy@gmail.com>
#
# GNU General Public License Usage
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

chopok <- function
(
    # number of task to solve
    task = 0,
    
    # seed
    seed = 14061987,
    
    # timing specification
    intervalCount = 3,                          # number of time intervals with different statistical parameters
    timeStamps = c(3600, 2 * 3600, 3 * 3600),   # time stamps (in seconds) marking right boundaries of intervals
    snapshotInterval1 = 5 * 60,                 # time in seconds per reviewing the number of dispatched/arrived passengers
    snapshotInterval2 = 1 * 60,                 # time in seconds per reviewing the number of dispatched/arrived passengers
    snapshotInterval3 = 10 * 60,                # time in seconds per reviewing the number of dispatched/arrived passengers
    
    # Station One parameters
    oLambda = c(1.2 / 5, 1 / 5, 0.8 / 5),       # arrival rates (per second) for corresponding intervals
    oCarInterval = 20,                          # interval between two departures
    oCarCapacity = 4,                           # capacity per departure
    
    # transfer parameters
    transferCoin = 0.5,                         # probability of single transfer to Station Two
    
    # Station Two parameters
    tLambda = 1 / 20,                           # arrival rate (per second)
    tCarInterval = 20,                          # interval between two departures
    tCarCapacity = 2                            # capacity per departure
) 
{
    set.seed(seed)
    
    ############################################################################
    
    # Station One main data structures (further used for statistics)
    oArrivalCount <- 0                          # accumulates all arrivals
    oTimeBetweenArrivals <- NULL                # vector of time slices (in seconds) between arrivals of passengers
    oDepartureLoads <- NULL                     # vector of car loads (0-oCarCapacity)
    oFullLoadCount <- 0                         # counts the number of loads, which are equal to the oCarCapacity
    oDispatchSnapshots <- NULL                  # vector of dispatched passengers per snapshotInterval1
    oArrivalSnapshots1 <- NULL                  # vector of arrived passengers per snapshotInterval1
    oArrivalSnapshots2 <- NULL                  # vector of arrived passengers per snapshotInterval2
    oWaitingTimes <- NULL                       # vector of waiting times (in seconds) in queue (one value for each dispatched passenger)
                                                # together with oQueuePointer represents the queue
                                                # pushing the queue means appending the starting time of waiting (next closest departure) to the end of oWaitingTimes
    oQueueSizeHistory <- NULL                   # vector of queue sizes (logged at every second)
    
    # Station One auxiliary data structures
    oTime <- 0                                  # global timer (incremented by 1 second)
    oTimeAccumulator <- 0                       # accumulator for time between two arrivals
    oDepartureTimer <- 0                        # timer (in seconds) triggering departures of cars (zeroed at oCarInterval)
    oDepartureCount <- 0                        # number of departures (incremented at every departure)
    oSnapshotTimer1 <- 0                        # timer (in seconds) triggering dispatch snapshots (zeroed at snapshotInterval1)
    oSnapshotTimer2 <- 0                        # timer (in seconds) triggering dispatch snapshots (zeroed at snapshotInterval2)
    oDispatchAccumulator <- 0                   # accumulator for loads of departured cars (zeroed at snapshotInterval1)
    oArrivalAccumulator1 <- 0                   # accumulator for arrivals of passengers (zeroed at snapshotInterval1)
    oArrivalAccumulator2 <- 0                   # accumulator for arrivals of passengers (zeroed at snapshotInterval2)
    oQueuePointer <- 1                          # points to the top of the queue (vectors are indexed from 1)
                                                # together with oWaitingTimes represents the queue 
                                                # popping the queue means:
                                                #   1) writing back the difference between pointed element of oWaitingTimes and current time
                                                #   2) incrementing the oQueuePointer
    oQueueSize <- 0                             # current size of queue (the number of elements in oWaitingTimes after the element pointed by oQueuePointer)
    
    ############################################################################
    
    # Station One simulation
    for (i in 1:intervalCount) 
    {
    	while (oTime <= timeStamps[i] | (i == intervalCount & oQueueSize > 0)) 
        {            
            # tick clocks
            oTime <- oTime + 1
            oDepartureTimer <- oDepartureTimer + 1
            oSnapshotTimer1 <- oSnapshotTimer1 + 1
            oSnapshotTimer2 <- oSnapshotTimer2 + 1
            
            # generate arrivals
            randomArrivalCount <- 0
            if (oTime <= timeStamps[intervalCount]) # is gate closed?
                randomArrivalCount = rpois(1, oLambda[i]) # Poisson random value gives the number of arrivals per second
            
            # accumulate arrivals from the beggining
            oArrivalCount <- oArrivalCount + randomArrivalCount
            
            # accumulate arrivals since the last snapshot
            oArrivalAccumulator1 <- oArrivalAccumulator1 + randomArrivalCount
            oArrivalAccumulator2 <- oArrivalAccumulator2 + randomArrivalCount
            
            # store the time between arrivals
            if (randomArrivalCount > 0) 
            {    # somebody arrived, log time from the last arrival
                oTimeBetweenArrivals <- c(oTimeBetweenArrivals, oTimeAccumulator) 
                oTimeAccumulator <- 0
            }
            else # nobody arrived, maybe next time...
                oTimeAccumulator <- oTimeAccumulator + 1 
            
            # insert arrived passengers into the queue
            if (randomArrivalCount >= 1)
            {
              for (j in 1:randomArrivalCount)
              {
                  # store the time of expected departure (closest possible in the future = starting of waiting) for every arrived passenger
                  oWaitingTimes <- c(oWaitingTimes, oDepartureCount * oCarInterval + oCarInterval)
                  oQueueSize <- oQueueSize + 1
              }
            }
    		
    		# car is departing
            if (oDepartureTimer == oCarInterval)
            {
                # log departure
                oDepartureCount <- oDepartureCount + 1
                
                # fill a car up to its capacity (depending on available passengers in the queue)
                load <- 0
                while (load < oCarCapacity & oQueueSize > 0) 
  		        {
  		            # store the waiting time of passenger from the top of the queue
                    oWaitingTimes[oQueuePointer] <- oTime - oWaitingTimes[oQueuePointer] # difference between current time and start of waiting
                    # remove passenger from the top of the queue
                    oQueuePointer <- oQueuePointer + 1
                    oQueueSize <- oQueueSize - 1
                    # insert passenger to the car
                    load <- load + 1
                }
                
                # log full load of the car
                if (load == oCarCapacity)
                    oFullLoadCount <- oFullLoadCount + 1
                    
                # accumulate number of dispatched passengers since the last snapshot
                oDispatchAccumulator <- oDispatchAccumulator + load
                
                # store departure
                oDepartureLoads <- c(oDepartureLoads, load)
                
  		        oDepartureTimer <- 0
            }
            
            # take snapshots
            if (oSnapshotTimer1 == snapshotInterval1)
            {
                oArrivalSnapshots1 <- c(oArrivalSnapshots1, oArrivalAccumulator1)
                oDispatchSnapshots <- c(oDispatchSnapshots, oDispatchAccumulator)
                oArrivalAccumulator1 <- 0
                oDispatchAccumulator <- 0
                oSnapshotTimer1 <- 0
            }
            if (oSnapshotTimer2 == snapshotInterval3)
            {
                oArrivalSnapshots2 <- c(oArrivalSnapshots2, oArrivalAccumulator2)
                oArrivalAccumulator2 <- 0
                oSnapshotTimer2 <- 0
            }
            
            # log queue size
            oQueueSizeHistory <- c(oQueueSizeHistory, oQueueSize)
    	}
    }
    
    ############################################################################
    
    # Station Two main data structures (further used for statistics)
    tArrivalCount <- 0                          # accumulates all arrivals
    tTransferredCount <- 0                      # accumulates transferred arrivals generated by tossing a coin
    tTimeBetweenArrivals <- NULL                # vector of time slices (in seconds) between arrivals of passengers
    tDepartureLoads <- NULL                     # vector of car loads (0-tCarCapacity)
    tFullLoadCount <- 0                         # counts the number of loads, which are equal to the tCarCapacity
    tDispatchSnapshots <- NULL                  # vector of dispatched passengers per snapshotInterval1
    tArrivalSnapshots <- NULL                   # vector of arrived passengers per snapshotInterval2
    tWaitingTimes <- NULL                       # vector of waiting times (in seconds) in queue (one for each dispatched passenger)
                                                # together with tQueuePointer represents the queue
                                                # pushing the queue means appending the starting time of waiting (next closest departure) to the end of tWaitingTimes
    tQueueSizeHistory <- NULL                   # vector of queue sizes (logged at every second)
    
    # Station Two auxiliary data structures
    tTime <- 0                                  # global timer (incremented by 1 second)
    tTimeAccumulator <- 0                       # accumulator for time between two arrivals
    tDepartureTimer <- 0                        # timer (in seconds) triggering departures of cars (zeroed at tCarInterval)
    tArrivalTimer <- 0                          # timer (in seconds) triggering arrivals of cars from Station One (zeroed at oCarInterval)
    tDepartureCount <- 0                        # number of departures (incremented at every departure)
    tSnapshotTimer1 <- 0                        # timer (in seconds) triggering dispatch snapshots (zeroed at snapshotInterval1)
    tSnapshotTimer2 <- 0                        # timer (in seconds) triggering dispatch snapshots (zeroed at snapshotInterval2)
    tDispatchAccumulator <- 0                   # accumulator for loads of departured cars (zeroed at snapshotInterval1)
    tArrivalAccumulator <- 0                    # accumulator for arrivals of passengers (zeroed at snapshotInterval2)
    tQueuePointer <- 1                          # points to the top of the queue (vectors are indexed from 1)
                                                # together with tWaitingTimes represents the queue 
                                                # popping the queue means:
                                                #   1) writing back the difference between pointed element of tWaitingTimes and current time
                                                #   2) incrementing the tQueuePointer
    tQueueSize <- 0                             # current size of queue (the number of elements in tWaitingTimes after the element pointed by tQueuePointer)
    tNextArrivalTime <- 0                       # time of the closest next arrival in the future
    tArrivalLoadsFromOnePointer <- 1            # pointer into oDepartureLoads vector simulated at Station One (vectors are indexed from 1)
                                                # elements of oDepartureLoads determines the number of arrivals exposed to the transfer decision
                                                # note that Station Two is delayed by the time, which is equal to the lift time of Station One
                                                # thus, output from Station One is directly the input for Station Two (without any need to shift vectors)
    
    ############################################################################
    
    # Station Two simulation
    # exponential random value gives the time between arrivals
    tNextArrivalTime <- rexp(1, tLambda) # schedule first arrival from environment (not a transfer from Station One)
  	while (tTime <= timeStamps[intervalCount] | tQueueSize > 0) 
    {
        # tick clocks
        tTime <- tTime + 1
        tArrivalTimer <- tArrivalTimer + 1
        tDepartureTimer <- tDepartureTimer + 1
        tSnapshotTimer1 <- tSnapshotTimer1 + 1
        tSnapshotTimer2 <- tSnapshotTimer2 + 1
        tNextArrivalTime <- tNextArrivalTime - 1 # future is closer
        
        # generate arrivals
        randomArrivalCount <- 0
        if (tTime <= timeStamps[intervalCount]) # is gate closed?
        {
            # is there a car arriving from Station One?
            if (tArrivalTimer == oCarInterval)
            {
                for (i in 1:oDepartureLoads[tArrivalLoadsFromOnePointer])
                {
                    # every passenger from Station One tosses a coin whether to transfer to Station Two or not
                    decision <- rbinom(1, 1, transferCoin)
                    randomArrivalCount <- randomArrivalCount + decision
                    tTransferredCount <- tTransferredCount + decision
                }
                tArrivalLoadsFromOnePointer <- tArrivalLoadsFromOnePointer + 1  # next car
                tArrivalTimer <- 0
            }
            
            # is there any passenger comming from environment?
            if (tNextArrivalTime <= 0)
            {
                randomArrivalCount <- randomArrivalCount + 1
                tNextArrivalTime <- rexp(1, tLambda) # schedule next one
            }
        }
        
        # accumulate arrivals from the beggining
        tArrivalCount <- tArrivalCount + randomArrivalCount
        
        # accumulate arrivals since the last snapshot
        tArrivalAccumulator <- tArrivalAccumulator + randomArrivalCount
        
        # store the time between arrivals
        if (randomArrivalCount > 0) 
        {    # somebody arrived, log time from the last arrival
            tTimeBetweenArrivals <- c(tTimeBetweenArrivals, tTimeAccumulator) 
            tTimeAccumulator <- 0
        }
        else # nobody arrived, maybe next time...
            tTimeAccumulator <- tTimeAccumulator + 1 
        
        # insert arrived passengers into the queue
        if (randomArrivalCount >= 1)
        {
            for (j in 1:randomArrivalCount)
            {
                # store the time of expected departure (closest possible in the future = starting of waiting) for every arrived passenger
                tWaitingTimes <- c(tWaitingTimes, tDepartureCount * tCarInterval + tCarInterval)
                tQueueSize <- tQueueSize + 1
            }
        }
  		
  		# car is departing
        if (tDepartureTimer == tCarInterval)
        {
            # log departure
            tDepartureCount <- tDepartureCount + 1
            
            # fill a car up to its capacity (depending on available passengers in the queue)
            load <- 0
            while (load < tCarCapacity & tQueueSize > 0) 
	        {
	            # store the waiting time of passenger from the top of the queue
                tWaitingTimes[tQueuePointer] <- tTime - tWaitingTimes[tQueuePointer] # difference between current time and start of waiting
                # remove passenger from the top of the queue
                tQueuePointer <- tQueuePointer + 1
                tQueueSize <- tQueueSize - 1
                # insert passenger to the car
                load <- load + 1
            }
            
            # log full load of car
            if (load == oCarCapacity)
                tFullLoadCount <- tFullLoadCount + 1
                
            # accumulate number of dispatched passengers since the last snapshot
            tDispatchAccumulator <- tDispatchAccumulator + load
            
            # store departure
            tDepartureLoads <- c(tDepartureLoads, load)
            
	        tDepartureTimer <- 0
        }
          
        # take snapshots
        if (tSnapshotTimer1 == snapshotInterval1)
        {
            tDispatchSnapshots <- c(tDispatchSnapshots, tDispatchAccumulator)
            tDispatchAccumulator <- 0
            tSnapshotTimer1 <- 0
        }
        if (tSnapshotTimer2 == snapshotInterval2)
        {
            tArrivalSnapshots <- c(tArrivalSnapshots, tArrivalAccumulator)
            tArrivalAccumulator <- 0
            tSnapshotTimer2 <- 0
        }
        
        # log queue size
        tQueueSizeHistory <- c(tQueueSizeHistory, tQueueSize)
  	}
        
    ############################################################################
    
    if (task == 1)
    {
        cat("Station One arrival count:", oArrivalCount, "\n")
        cat("Station Two arrival count:", tArrivalCount, "\n")
        cat("P[passenger at Station Two is from Station One] =", tTransferredCount / tArrivalCount, "\n")
    }    
    else if (task == 2)
    {
        cat("P[full load at Station One] =", oFullLoadCount / length(oDepartureLoads), "\n")
        cat("mean[load at Station One] =", mean(oDepartureLoads), "\n")
        cat("var[load at Station One] =", var(oDepartureLoads), "\n")
        prop.test(oFullLoadCount, length(oDepartureLoads), oFullLoadCount / length(oDepartureLoads))
    }
    else if (task == 3)
    {
        dispatchSnapshots <- NULL
        if (length(oDispatchSnapshots) <= length(tDispatchSnapshots))
        {
            for (i in 1:length(oDispatchSnapshots))
                dispatchSnapshots <- c(dispatchSnapshots, oDispatchSnapshots[i] + tDispatchSnapshots[i])
        }
        else
        {
            for (i in 1:length(tDispatchSnapshots))
                dispatchSnapshots <- c(dispatchSnapshots, oDispatchSnapshots[i] + tDispatchSnapshots[i])
        }
        
        h <- hist(dispatchSnapshots)
        xhist <- c(min(h$breaks), h$breaks)
        yhist <- c(0, h$density, 0)
        xnorm <- seq(min(dispatchSnapshots), max(dispatchSnapshots), length = 40)
        ynorm <- dnorm(xnorm, mean(dispatchSnapshots), sd(dispatchSnapshots))
        plot(xhist, yhist, type = "s", ylim = c(0, max(yhist, ynorm)), xlab = paste("Dispatched passengers per", snapshotInterval1 / 60, "minute(s)"), ylab = "Density")
        lines(xnorm, ynorm, col = "blue")
        
        shapiro.test(dispatchSnapshots)
    }
    else if (task == 4)
    {
        h <- hist(oArrivalSnapshots1)
        xhist <- c(min(h$breaks), h$breaks)
        yhist <- c(0, h$density, 0)
        xnorm <- seq(min(oArrivalSnapshots1), max(oArrivalSnapshots1), length = 40)
        ynorm <- dnorm(xnorm, mean(oArrivalSnapshots1), sd(oArrivalSnapshots1))
        plot(xhist, yhist, type = "s", ylim = c(0, max(yhist, ynorm)), xlab = paste("Arrivals per", snapshotInterval1 / 60, "minute(s) at Station One"), ylab = "Density")
        lines(xnorm, ynorm, col = "blue")
        
        shapiro.test(oArrivalSnapshots1)
    }
    else if (task == 5)
    {
        cat("mean[waiting time at Station One] =", mean(oWaitingTimes), "\n")
        cat("mean[waiting time at Station Two] =", mean(tWaitingTimes), "\n")
    }
    else if (task == 6 | task == 61)
    {
        h <- hist(oTimeBetweenArrivals)
        xhist <- c(min(h$breaks), h$breaks)
        yhist <- c(0, h$density, 0)
        xexp <- seq(min(oTimeBetweenArrivals), max(oTimeBetweenArrivals), length = 40)
        yexp <- dexp(xexp, 1 / mean(oTimeBetweenArrivals))
        plot(xhist, yhist, type = "s", ylim = c(0, max(yhist, yexp)), xlab = "Time between passenger arrivals at Station One", ylab = "Density")
        lines(xexp, yexp, col = "blue")
        
        cat("lambda =", 1 / mean(oTimeBetweenArrivals), "(arrived passengers per second at Station One)", "\n")
    }
    else if (task == 62)
    {
        h <- hist(tArrivalSnapshots)
        xhist <- c(min(h$breaks), h$breaks)
        yhist <- c(0, h$density, 0)
        xpois <- xhist
        ypois <- dpois(xpois, mean(tArrivalSnapshots))
        plot(xhist, yhist, type = "s", ylim = c(0, max(yhist, ypois)), xlab = paste("Arrived passengers per", snapshotInterval2 / 60, "minute(s) at Station Two"), ylab = "Density")
        lines(xpois, ypois, col = "blue")
        
        cat("lambda =", mean(tArrivalSnapshots), "(arrived passengers per", snapshotInterval2 / 60, "minute(s) at Station Two)", "\n")
    }
    else if (task == 7 | task == 71)
    {
        plot(1:length(oQueueSizeHistory), oQueueSizeHistory, type = "l", xlab = "Time (seconds)", ylab = "Queue size at Station One")
    }
    else if (task == 72)
    {
        plot(1:length(tQueueSizeHistory), tQueueSizeHistory, type = "l", xlab = "Time (seconds)", ylab = "Queue size at Station Two")
    }
    else if (task == 8)
    {
        cat("Station One", "\n")
        cat("Car interval =", oCarInterval, "\n")
        cat("Car capacity =", oCarCapacity, "\n")
        cat("Snapshot interval =", snapshotInterval3 / 60, "minutes", "\n")
        cat("Percentages of arrivals over maximum dispatch rate of the station:", "\n")
        cat(round(((oArrivalSnapshots2 / (snapshotInterval3 / oCarInterval * oCarCapacity)) * 100), 0), "\n")
    }
    else if (task == 9)
    {
        cat("Station One arrival count:", oArrivalCount, "\n")
    }
    else
    {
        cat("Chopok", "\n")
        cat("Ski lift statistic simulation", "\n")
        cat("Copyright (c) 2010 Petr Koupy <petr.koupy@gmail.com>", "\n")
        cat("Usage:", "\n")
        cat("\tchopok(task, seed)", "\n")
        cat("\tavailable tasks: 1, 2, 3, 4, 5, 61, 62, 71, 72, 8, 9", "\n")
        cat("Note: Further parametrization possible - see source by typing 'chopok'.", "\n")
    }
}
